// SPDX-FileCopyrightText: 2022 Matteo Settenvini <matteo.settenvini@montecristosoftware.eu>
// SPDX-License-Identifier: GPL-3.0-or-later

use {
    crate::helpers::{records_to_gaih_addr, records_to_hostent, set_if_valid, DnsResult},
    crate::nss_bindings::{gaih_addrtuple, nss_status, HErrno},
    crate::policy_checker::{PolicyChecker as _, POLICY_CHECKER},
    libc::{c_char, c_int, hostent, size_t},
    nix::errno,
    nix::errno::Errno,
    std::ffi::CStr,
    trust_dns_proto::rr::record_type::RecordType,
    trust_dns_proto::xfer::dns_request::DnsRequestOptions,
    trust_dns_resolver::{lookup::Lookup, lookup_ip::LookupIp},
};

// See https://support.mozilla.org/en-US/kb/configuring-networks-disable-dns-over-https
const CANARY_HOSTNAME: &str = "use-application-dns.net";

pub struct Args {
    pub name: *const c_char,
    pub family: c_int,
    pub buffer: *mut c_char,
    pub buflen: size_t,
    pub errnop: *mut Errno,
    pub h_errnop: *mut HErrno,
    pub ttlp: *mut i32,
    pub result: Result,
}

pub enum Result {
    V3(*mut hostent),
    V4(*mut *mut gaih_addrtuple),
}

pub async unsafe fn with(args: &mut Args) -> nss_status {
    set_if_valid(args.errnop, errno::from_i32(0));
    set_if_valid(args.h_errnop, HErrno::Success);

    match POLICY_CHECKER.resolver(None) {
        Ok(None) => {
            // no restrictions for user, the next NSS module will decide
            nss_status::NSS_STATUS_NOTFOUND
        }
        Ok(Some(resolver)) => {
            let name = match CStr::from_ptr(args.name).to_str() {
                Ok(name) => name,
                Err(_) => {
                    set_if_valid(args.errnop, Errno::EINVAL);
                    set_if_valid(args.h_errnop, HErrno::Internal);
                    return nss_status::NSS_STATUS_TRYAGAIN;
                }
            };

            // disable application-based DNS for those applications
            // (notably, Firefox) that support it
            if name == CANARY_HOSTNAME {
                set_if_valid(args.h_errnop, HErrno::HostNotFound);
                return nss_status::NSS_STATUS_SUCCESS;
            }

            let lookup: std::result::Result<Lookup, _> = match args.family {
                libc::AF_UNSPEC => resolver.lookup_ip(name).await.map(LookupIp::into),
                libc::AF_INET => {
                    resolver
                        .lookup(name, RecordType::A, DnsRequestOptions::default())
                        .await
                }
                libc::AF_INET6 => {
                    resolver
                        .lookup(name, RecordType::AAAA, DnsRequestOptions::default())
                        .await
                }
                _ => return nss_status::NSS_STATUS_NOTFOUND,
            };

            match lookup {
                Ok(result) => prepare_response(args, result),
                Err(err) => {
                    log::warn!("{}", err);
                    nss_status::NSS_STATUS_UNAVAIL
                }
            }
        }
        Err(err) => {
            log::error!("{}", err);
            nss_status::NSS_STATUS_UNAVAIL
        }
    }
}

unsafe fn prepare_response(
    args: &mut Args,
    lookup: trust_dns_resolver::lookup::Lookup,
) -> nss_status {
    if lookup.iter().peekable().peek().is_none() {
        set_if_valid(args.h_errnop, HErrno::HostNotFound);
        return nss_status::NSS_STATUS_SUCCESS;
    }

    let ttl = lookup
        .valid_until()
        .duration_since(std::time::Instant::now())
        .as_secs();
    set_if_valid(
        args.ttlp,
        if ttl < (i32::MAX as u64) {
            ttl as i32
        } else {
            i32::MAX
        },
    );

    let buf = std::slice::from_raw_parts_mut(args.buffer as *mut u8, args.buflen);
    let ret = match &mut args.result {
        Result::V3(hostent) => {
            if hostent.is_null() {
                set_if_valid(args.errnop, Errno::EINVAL);
                set_if_valid(args.h_errnop, HErrno::Internal);
                return nss_status::NSS_STATUS_TRYAGAIN;
            }

            match records_to_hostent(lookup.into(), &mut **hostent, buf) {
                Ok(DnsResult::Found) => nss_status::NSS_STATUS_SUCCESS,
                Ok(DnsResult::NxDomain) => {
                    set_if_valid(args.h_errnop, HErrno::HostNotFound);
                    nss_status::NSS_STATUS_SUCCESS
                }
                Err(err) => {
                    set_if_valid(
                        args.errnop,
                        err.raw_os_error()
                            .map(Errno::from_i32)
                            .unwrap_or(Errno::EAGAIN),
                    );
                    set_if_valid(args.h_errnop, HErrno::Internal);
                    nss_status::NSS_STATUS_TRYAGAIN
                }
            }
        }
        Result::V4(pat) => {
            if pat.is_null() {
                set_if_valid(args.errnop, Errno::EINVAL);
                set_if_valid(args.h_errnop, HErrno::Internal);
                return nss_status::NSS_STATUS_TRYAGAIN;
            }

            match records_to_gaih_addr(lookup.into(), buf) {
                Ok(addrs) => {
                    // DEBUG: eprintln!("{:?} => {:?}", addrs, *addrs);
                    **pat = addrs;
                    nss_status::NSS_STATUS_SUCCESS
                }
                Err(err) => {
                    set_if_valid(
                        args.errnop,
                        err.raw_os_error()
                            .map(Errno::from_i32)
                            .unwrap_or(Errno::EAGAIN),
                    );
                    set_if_valid(args.h_errnop, HErrno::Internal);
                    nss_status::NSS_STATUS_TRYAGAIN
                }
            }
        }
    };

    ret
}
