// SPDX-FileCopyrightText: 2022 Matteo Settenvini <matteo.settenvini@montecristosoftware.eu>
// SPDX-License-Identifier: GPL-3.0-or-later

use {
    anyhow::Result,
    nix::unistd::{getuid, Uid},
    once_cell::sync::Lazy,
    std::collections::HashMap,
    std::net::{IpAddr, Ipv4Addr, Ipv6Addr, SocketAddr},
    std::sync::{Arc, RwLock},
    trust_dns_resolver::config as dns_config,
    trust_dns_resolver::TokioAsyncResolver,
};

pub type Restrictions<'a> = &'a [IpAddr];

pub trait PolicyChecker {
    fn resolver(&self, user: Option<Uid>) -> Result<Option<Arc<TokioAsyncResolver>>>;
}

pub static POLICY_CHECKER: Lazy<MalcontentPolicyChecker> =
    Lazy::new(|| MalcontentPolicyChecker::new());

static CLOUDFLARE_PARENTALCONTROL_ADDRS: Lazy<Vec<IpAddr>> = Lazy::new(|| {
    vec![
        IpAddr::V4(Ipv4Addr::new(1, 1, 1, 3)),
        IpAddr::V4(Ipv4Addr::new(1, 0, 0, 3)),
        IpAddr::V6(Ipv6Addr::new(2606, 4700, 4700, 0, 0, 0, 0, 1113)),
        IpAddr::V6(Ipv6Addr::new(2606, 4700, 4700, 0, 0, 0, 0, 1003)),
    ]
});

// TODO: accept notifications about config changes
pub struct MalcontentPolicyChecker {
    resolvers: RwLock<HashMap<Uid, Option<Arc<TokioAsyncResolver>>>>,
}

impl MalcontentPolicyChecker {
    pub fn new() -> Self {
        Self {
            resolvers: RwLock::new(HashMap::new()),
        }
    }

    fn restrictions<'a>(&'a self, user: Uid) -> Result<Option<Restrictions<'a>>> {
        if user.is_root() {
            return Ok(None);
        };

        // TODO: for now, hardcoded DNS records, later go
        // through D-Bus to malcontent and ask about user.

        Ok(Some(CLOUDFLARE_PARENTALCONTROL_ADDRS.as_slice()))
    }
}

impl PolicyChecker for MalcontentPolicyChecker {
    fn resolver(&self, user: Option<Uid>) -> Result<Option<Arc<TokioAsyncResolver>>> {
        let user = user.unwrap_or_else(|| getuid());
        {
            let ro_resolvers = self.resolvers.read().unwrap();
            if let Some(resolver) = ro_resolvers.get(&user) {
                return Ok(resolver.clone());
            }
        }

        {
            let mut rw_resolvers = self.resolvers.write().unwrap();
            if rw_resolvers.contains_key(&user) {
                return self.resolver(Some(user));
            }

            let resolver = match self.restrictions(user)? {
                Some(addrs) => {
                    let resolver_config =
                        addrs
                            .iter()
                            .fold(dns_config::ResolverConfig::new(), |mut config, addr| {
                                config.add_name_server(dns_config::NameServerConfig {
                                    socket_addr: SocketAddr::new(*addr, 53),
                                    protocol: dns_config::Protocol::Udp,
                                    tls_dns_name: None,
                                    trust_nx_responses: true,
                                    bind_addr: None,
                                });
                                config
                            });

                    let resolver = TokioAsyncResolver::tokio(
                        resolver_config,
                        dns_config::ResolverOpts::default(),
                    )?;
                    Some(Arc::new(resolver))
                }
                None => None,
            };
            rw_resolvers.insert(user, resolver);
        }

        self.resolver(Some(user))
    }
}
